'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.multi = exports.placeholder = exports.backdrop = exports.selection = exports.firstLine = exports.firstLetter = exports.before = exports.after = exports.nthOfType = exports.nthLastOfType = exports.nthLastChild = exports.nthChild = exports.not = exports.lang = exports.dir = exports.visited = exports.valid = exports.target = exports.scope = exports.root = exports.right = exports.required = exports.readWrite = exports.readOnly = exports.outOfRange = exports.optional = exports.onlyOfType = exports.onlyChild = exports.link = exports.left = exports.lastOfType = exports.lastChild = exports.invalid = exports.inRange = exports.indeterminate = exports.hover = exports.focus = exports.fullscreen = exports.firstOfType = exports.firstChild = exports.first = exports._default = exports.enabled = exports.empty = exports.disabled = exports.checked = exports.any = exports.active = undefined;

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj; };

exports.simulations = simulations;
exports.simulate = simulate;
exports.cssLabels = cssLabels;
exports.speedy = speedy;
exports.appendSheetRule = appendSheetRule;
exports.flush = flush;
exports.remove = remove;
exports.idFor = idFor;
exports.add = add;
exports.style = style;
exports.select = select;
exports.keyed = keyed;
exports.merge = merge;
exports.media = media;
exports.trackMediaQueryLabels = trackMediaQueryLabels;
exports.fontFace = fontFace;
exports.keyframes = keyframes;
exports.cssFor = cssFor;
exports.attribsFor = attribsFor;
exports.renderStatic = renderStatic;
exports.renderStaticOptimized = renderStaticOptimized;
exports.rehydrate = rehydrate;

var _hash = require('./hash');

var _hash2 = _interopRequireDefault(_hash);

var _autoprefix = require('./autoprefix');

var _autoprefix2 = _interopRequireDefault(_autoprefix);

var _CSSPropertyOperations = require('./CSSPropertyOperations');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } } // first we import some helpers 
// hashes a string to something 'unique'

// yurgh must get back to this 
// import prefixAll from 'inline-style-prefixer/static'   // adds vendor prefixes to styles 
// import  Prefix  from 'inline-style-prefixer'
// let prefixer = new Prefix({ userAgent: navigator.userAgent })


// we've used browserify to extract react's CSSPropertyOperations module and it's deps into ./CSSPropertyOperations 


// converts a js style object to css markup
// todo - rewrite this yourself, save a kb or two 

// define some constants 
var isBrowser = typeof document !== 'undefined';
var isDev = function (x) {
  return x === 'development' || !x;
}(process.env.NODE_ENV);
var isTest = process.env.NODE_ENV === 'test';

// a useful utility for quickly tapping objects. use with the :: operator 
// {x: 1}::log()
// [5, 12, 90]::log().filter(x => x%5)::log()
function log(msg) {
  //eslint-disable-line no-unused-vars
  console.log(msg || this); //eslint-disable-line no-console
  return this;
}

// takes a string, converts to lowercase, strips out nonalphanumeric.
function simple(str) {
  return str.toLowerCase().replace(/[^a-z0-9]/g, '');
}

/**** simulations  ****/

// a flag to enable simulation meta tags on dom nodes 
// defaults to true in dev mode. recommend *not* to 
// toggle often. 
var canSimulate = isDev;

// we use these flags for issuing warnings when simulate is called 
// in prod / in incorrect order 
var warned1 = false,
    warned2 = false;

// toggles simulation activity. shouldn't be needed in most cases 
function simulations() {
  var bool = arguments.length <= 0 || arguments[0] === undefined ? true : arguments[0];

  canSimulate = !!bool;
}

// use this on dom nodes to 'simulate' pseudoclasses
// <div {...hover({ color: 'red' })} {...simulate('hover', 'visited')}>...</div>
// you can even send in some weird ones, as long as it's in simple format 
// and matches an existing rule on the element 
// eg simulate('nthChild2', ':hover:active') etc 
function simulate() {
  if (!canSimulate) {
    if (!warned1) {
      console.warn('can\'t simulate without once calling simulations(true)'); //eslint-disable-line no-console
      warned1 = true;
    }
    if (!isDev && !isTest && !warned2) {
      console.warn('don\'t use simulation outside dev'); //eslint-disable-line no-console
      warned2 = true;
    }
    return {};
  }

  for (var _len = arguments.length, pseudos = Array(_len), _key = 0; _key < _len; _key++) {
    pseudos[_key] = arguments[_key];
  }

  return pseudos.reduce(function (o, p) {
    return o['data-simulate-' + simple(p)] = '', o;
  }, {});
}

/**** labels ****/
// toggle for debug labels. 
// shouldn't *have* to mess with this manually
var hasLabels = false; // isDev

function cssLabels(bool) {
  hasLabels = !!bool;
}

/**** stylesheet ****/

// these here are our main 'mutable' references
var cache = {},
    // stores all the registered styles. most important, for such a small name.  
styleTag = void 0,
    // reference to the <style> tag, if in browser 
styleSheet = void 0,
    // reference to the styleSheet object, either native on browser / polyfilled on server 
keyIndices = {}; // avoid scanning when inserting rules 


function injectStyleSheet() {
  if (isBrowser) {
    // this section is just weird alchemy I found online off many sources 
    // it checks to see if the tag exists; creates an empty one if not 
    styleTag = document.getElementById('_css_');
    if (!styleTag) {
      styleTag = document.createElement('style');
      styleTag.id = styleTag.id || '_css_';
      styleTag.appendChild(document.createTextNode(''));
      (document.head || document.getElementsByTagName('head')[0]).appendChild(styleTag);
    }
    // this weirdness brought to you by firefox 
    styleSheet = [].concat(_toConsumableArray(document.styleSheets)).filter(function (x) {
      return x.ownerNode === styleTag;
    })[0];
  } else {
    // server side 'polyfill'. just enough behavior to be useful.
    styleSheet = {
      cssRules: [],
      deleteRule: function deleteRule(index) {
        styleSheet.cssRules = [].concat(_toConsumableArray(styleSheet.cssRules.slice(0, index)), _toConsumableArray(styleSheet.cssRules.slice(index + 1)));
      },
      insertRule: function insertRule(rule, index) {
        // enough 'spec compliance' to be able to extract the rules later  
        // in other words, just the cssText field 
        styleSheet.cssRules = [].concat(_toConsumableArray(styleSheet.cssRules.slice(0, index)), [{ cssText: rule }], _toConsumableArray(styleSheet.cssRules.slice(index)));
      }
    };
  }
}

/**************** LIFTOFF IN 3... 2... 1... ****************/
injectStyleSheet();
/****************      TO THE MOOOOOOON     ****************/

// a flag to use stylesheet.insertrule 
// the big drawback here is that the css won't be editable in devtools
var isSpeedy = !isDev && !isTest; // only in prod mode does it make 'sense' 

function speedy() {
  var bool = arguments.length <= 0 || arguments[0] === undefined ? true : arguments[0];

  // we don't let you change isSpeedy if you've already made a modification to the stylesheet
  if (bool !== isSpeedy && Object.keys(cache).length !== 0) {
    console.error('cannot change speedy setting after appending styles in a different mode'); //eslint-disable-line no-console
    return;
  }
  isSpeedy = !!bool;
}

function inlineInsertRule(rule) {
  var index = arguments.length <= 1 || arguments[1] === undefined ? styleSheet.cssRules.length : arguments[1];

  // this weirdness for perf, and chrome's weird bug 
  // https://stackoverflow.com/questions/20007992/chrome-suddenly-stopped-accepting-insertrule

  try {
    styleSheet.insertRule(rule, index);
  } catch (e) {
    if (isDev) {
      // might need beter dx for this 
      console.warn('whoops, illegal rule inserted', rule); //eslint-disable-line no-console
    }
  }
}

// adds a css rule to the sheet. only used 'internally'. 
function appendSheetRule(rule, index) {
  // todo - tests 

  // more browser weirdness. I don't even know
  if (styleTag && styleTag.styleSheet) {
    styleTag.styleSheet.cssText += rule;
  } else {
    if (isBrowser) {
      if (isSpeedy && styleSheet.insertRule) {
        inlineInsertRule(rule, index);
      } else {
        styleTag.appendChild(document.createTextNode(rule));
        // todo - more efficent here please 
        if (!isSpeedy) {
          // sighhh
          styleSheet = [].concat(_toConsumableArray(document.styleSheets)).filter(function (x) {
            return x.ownerNode === styleTag;
          })[0];
        }
      }
    } else {
      // server side is pretty simple 
      styleSheet.insertRule(rule, styleSheet.cssRules.length);
    }
  }
}

// clears out the cache and empties the stylesheet
// best for tests, though there might be some value for SSR. 
function flush() {
  // todo - tests 
  cache = {};
  // todo backward compat (styleTag.styleSheet.cssText?)
  if (isBrowser) {
    styleTag && styleTag.parentNode.removeChild(styleTag);
    styleTag = null;
    // todo - look for remnants in document.styleSheets
    injectStyleSheet();
  } else {
    // simpler on server 
    styleSheet.cssRules = [];
  }
}

function remove() {
  // todo
  // remove rule
  throw new Error('this is not tested or anything yet! beware!'); //eslint-disable-line no-console

  // let id = o[Object.keys(o)[0]]
  // let i = sheet.rules.indexOf(x => x.selectorText === selector(id, cache[id].type))
  // sheet.deleteRule(i)
  // delete cache[id]
}

// now, some functions to help deal with styles / rules 

// generates a hash for (type, style)
function styleHash(type, style) {
  // todo - default type = '_'. this changes all the hashes and will break tests, so do later 
  // make sure type exists
  // make sure obj is style-like?
  return (0, _hash2.default)(type + Object.keys(style).reduce(function (str, k) {
    return str + k + style[k];
  }, '')).toString(36);
}

// helper to hack around isp's array format 
function prefixes(style) {
  return (0, _autoprefix2.default)(style);
}

// generates a css selector for (id, type)
function selector(id, type) {
  // id should exist
  var isFullSelector = type && type[0] === '$';
  var cssType = type === '_' ? '' : type[0] === '$' ? type.slice(1) : ':' + type;
  var suffix = void 0;

  if (isFullSelector) {
    suffix = cssType.split(',').map(function (x) {
      return '[data-css-' + id + ']' + x;
    }).join(',');
  } else {
    suffix = '[data-css-' + id + ']' + cssType;
  }

  if (canSimulate && type !== '_' && !isFullSelector && cssType[0] === ':') {
    // todo - work with pseudo selector  on full selector at least 
    suffix += ', [data-css-' + id + '][data-simulate-' + simple(type) + ']';
  }
  return suffix;
}

// ... which is them used to generate css rules 
function cssrule(id, type, style) {
  return selector(id, type) + '{ ' + (0, _CSSPropertyOperations.createMarkupForStyles)(prefixes(style)) + ' } ';
}

// given a rule {data-css-id: ''}, checks if it's a valid, registered id
// returns the id 
function idFor(rule) {
  // todo - weak map hash for this?
  if (Object.keys(rule).length !== 1) throw new Error('not a rule');
  var regex = /data\-css\-([a-zA-Z0-9]+)/;
  var match = regex.exec(Object.keys(rule)[0]);
  if (!match) throw new Error('not a rule');
  return match[1];
}

// checks if a rule is registered
function isRule(rule) {
  try {
    var id = idFor(rule);
    return id && cache[id];
  } catch (e) {
    return false;
  }
}

// a generic rule creator/insertor 
function add() {
  var type = arguments.length <= 0 || arguments[0] === undefined ? '_' : arguments[0];
  var style = arguments[1];
  var key = arguments[2];

  var id = key || styleHash(type, style),
      // generate a hash based on type/style, use this to 'id' the rule everywhere 
  label = '',
      keyIndex = -1;

  if (!cache[id] || key) {
    if (key) {
      // if the key already exists, delete it 
      keyIndex = keyIndices[key];
      if (keyIndex >= 0) {
        //remove rule
        if (isSpeedy || !isBrowser) {
          styleSheet.deleteRule(keyIndex);
        } else {
          styleTag.removeChild(styleTag.childNodes[keyIndex + 1]); // the +1 to account for the blank node we added
          // reassign stylesheet, because firefox is weird 
          styleSheet = [].concat(_toConsumableArray(document.styleSheets)).filter(function (x) {
            return x.ownerNode === styleTag;
          })[0];
        }
      }
    }

    // add rule to sheet, update cache. easy!
    if (key) {
      appendSheetRule(cssrule(id, type, style), keyIndex);
      keyIndices[key] = keyIndex || styleSheet.cssRules.length - 1;
    } else {
      appendSheetRule(cssrule(id, type, style));
    }

    cache[id] = { type: type, style: style, id: id };
  }
  if (hasLabels) {
    // adds a debug label 
    label = style.label || (type !== '_' ? ':' + type : '');
  }

  return _defineProperty({}, 'data-css-' + id, label);
}

// with those in place, we can now define user-friendly functions for 
// defining styles on nodes 

// first up, what will probably be most commonly used.
// defines some css 'directly' on the node it's applied on
function style(obj) {
  return add(undefined, obj);
}

// alllllll the pseudoclasses
// todo - autogenerate this by scraping MDN
var active = exports.active = function active(x) {
  return add('active', x);
};
var any = exports.any = function any(x) {
  return add('any', x);
};
var checked = exports.checked = function checked(x) {
  return add('checked', x);
};
var disabled = exports.disabled = function disabled(x) {
  return add('disabled', x);
};
var empty = exports.empty = function empty(x) {
  return add('empty', x);
};
var enabled = exports.enabled = function enabled(x) {
  return add('enabled', x);
};
var _default = exports._default = function _default(x) {
  return add('default', x);
}; // note '_default' name 
var first = exports.first = function first(x) {
  return add('first', x);
};
var firstChild = exports.firstChild = function firstChild(x) {
  return add('first-child', x);
};
var firstOfType = exports.firstOfType = function firstOfType(x) {
  return add('first-of-type', x);
};
var fullscreen = exports.fullscreen = function fullscreen(x) {
  return add('fullscreen', x);
};
var focus = exports.focus = function focus(x) {
  return add('focus', x);
};
var hover = exports.hover = function hover(x) {
  return add('hover', x);
};
var indeterminate = exports.indeterminate = function indeterminate(x) {
  return add('indeterminate', x);
};
var inRange = exports.inRange = function inRange(x) {
  return add('in-range', x);
};
var invalid = exports.invalid = function invalid(x) {
  return add('invalid', x);
};
var lastChild = exports.lastChild = function lastChild(x) {
  return add('last-child', x);
};
var lastOfType = exports.lastOfType = function lastOfType(x) {
  return add('last-of-type', x);
};
var left = exports.left = function left(x) {
  return add('left', x);
};
var link = exports.link = function link(x) {
  return add('link', x);
};
var onlyChild = exports.onlyChild = function onlyChild(x) {
  return add('only-child', x);
};
var onlyOfType = exports.onlyOfType = function onlyOfType(x) {
  return add('only-of-type', x);
};
var optional = exports.optional = function optional(x) {
  return add('optional', x);
};
var outOfRange = exports.outOfRange = function outOfRange(x) {
  return add('out-of-range', x);
};
var readOnly = exports.readOnly = function readOnly(x) {
  return add('read-only', x);
};
var readWrite = exports.readWrite = function readWrite(x) {
  return add('read-write', x);
};
var required = exports.required = function required(x) {
  return add('required', x);
};
var right = exports.right = function right(x) {
  return add('right', x);
};
var root = exports.root = function root(x) {
  return add('root', x);
};
var scope = exports.scope = function scope(x) {
  return add('scope', x);
};
var target = exports.target = function target(x) {
  return add('target', x);
};
var valid = exports.valid = function valid(x) {
  return add('valid', x);
};
var visited = exports.visited = function visited(x) {
  return add('visited', x);
};

// parameterized pseudoclasses
var dir = exports.dir = function dir(p, x) {
  return add('dir(' + p + ')', x);
};
var lang = exports.lang = function lang(p, x) {
  return add('lang(' + p + ')', x);
};
var not = exports.not = function not(p, x) {
  return add('not(' + p + ')', x);
};
var nthChild = exports.nthChild = function nthChild(p, x) {
  return add('nth-child(' + p + ')', x);
};
var nthLastChild = exports.nthLastChild = function nthLastChild(p, x) {
  return add('nth-last-child(' + p + ')', x);
};
var nthLastOfType = exports.nthLastOfType = function nthLastOfType(p, x) {
  return add('nth-last-of-type(' + p + ')', x);
};
var nthOfType = exports.nthOfType = function nthOfType(p, x) {
  return add('nth-of-type(' + p + ')', x);
};

// pseudoelements
var after = exports.after = function after(x) {
  return add(':after', x);
};
var before = exports.before = function before(x) {
  return add(':before', x);
};
var firstLetter = exports.firstLetter = function firstLetter(x) {
  return add(':first-letter', x);
};
var firstLine = exports.firstLine = function firstLine(x) {
  return add(':first-line', x);
};
var selection = exports.selection = function selection(x) {
  return add(':selection', x);
};
var backdrop = exports.backdrop = function backdrop(x) {
  return add(':backdrop', x);
};
var placeholder = exports.placeholder = function placeholder(x) {
  return add(':placeholder', x);
};

// when you need multiple pseudoclasses in a single selector
// eg x:hover:visited for when hovering over visited elements 
var multi = exports.multi = add;

// unique feature 
// when you need to define 'real' css (whatever that may be)
// https://twitter.com/threepointone/status/756585907877273600
// https://twitter.com/threepointone/status/756986938033254400
function select(selector, style) {
  return add('$' + selector, style); // signalling ahead that this is a plain selector 
}

// unique feature 
// use for advanced perf/animations/whatnot 
// instead of overwriting, it replaces the rule in the stylesheet
function keyed(key, type, style) {
  // todo - accept a style/rule? unlcear. 
  if (typeof key !== 'string') {
    throw new Error('whoops, did you forget a key?');
  }
  if (!style && (typeof type === 'undefined' ? 'undefined' : _typeof(type)) === 'object') {
    style = type;
    type = undefined;
  }
  // should be able to pass a merged rule etc too 
  // maybe ...styles as well?
  return add(type, style, key);
}

// we define a function to 'merge' styles together.
// backstory - because of a browser quirk, multiple styles are applied in the order they're 
// defined the stylesheet, not in the order of application 
// in most cases, thsi won't case an issue UNTIL IT DOES 
// instead, use merge() to merge styles,
// with latter styles gaining precedence over former ones 
function merge() {
  var labels = [],
      mergeLabel = void 0,
      styleBag = {},
      mediaBag = {};

  for (var _len2 = arguments.length, rules = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
    rules[_key2] = arguments[_key2];
  }

  rules.forEach(function (rule, i) {
    // optionally send a string as first argumnet to 'label' this merged rule  
    if (i === 0 && typeof rule === 'string') {
      mergeLabel = rule;
      // bail early
      return;
    }
    if (isRule(rule)) {
      // it's a rule!

      var _id = idFor(rule);

      if (cache[_id].bag) {
        var _ret = function () {
          // merged rule 

          var _cache$_id = cache[_id];
          var bag = _cache$_id.bag;
          var label = _cache$_id.label;
          var media = _cache$_id.media;

          Object.keys(bag).forEach(function (type) {
            styleBag[type] = _extends({}, styleBag[type] || {}, bag[type]);
          });
          // if there's a media bag, merge those in 
          if (media) {
            Object.keys(media).forEach(function (expr) {
              mediaBag[expr] = mediaBag[expr] || {};
              Object.keys(media[expr]).forEach(function (type) {
                // mediaBag[expr][type] = mediaBag[expr][type] || {}
                mediaBag[expr][type] = _extends({}, mediaBag[expr][type] || {}, media[expr][type]);
              });
            });
          }

          hasLabels && labels.push('[' + label + ']');
          return {
            v: void 0
          };
          // that was fairly straightforward
        }();

        if ((typeof _ret === 'undefined' ? 'undefined' : _typeof(_ret)) === "object") return _ret.v;
      }

      if (cache[_id].expr) {
        var _ret2 = function () {
          // media rule
          var _cache$_id2 = cache[_id];
          var expr = _cache$_id2.expr;
          var label = _cache$_id2.label;
          var rule = _cache$_id2.rule;
          var style = _cache$_id2.style;

          mediaBag[expr] = mediaBag[expr] || {};
          if (rule) {
            var iid = idFor(rule);
            if (cache[iid].bag) {
              (function () {
                // if merged rule, merge it's bag into stylebag 
                // we won't expect a mediabag in this merged rule, because it would have thrown in media (phew)

                var bag = cache[iid].bag;

                Object.keys(bag).forEach(function (type) {
                  mediaBag[expr][type] = _extends({}, mediaBag[expr][type] || {}, bag[type]);
                });
              })();
            } else {
              var _cache$iid = cache[iid];
              var type = _cache$iid.type;
              var _style = _cache$iid.style;

              mediaBag[expr][type] = _extends({}, mediaBag[expr][type] || {}, _style);
            }
          } else {
            mediaBag[expr]._ = _extends({}, mediaBag[expr]._ || {}, style);
          }

          // mediaBag[expr].push(rule)
          hasLabels && labels.push(label);
          return {
            v: void 0
          };

          // throw new Error('cannot merge a media rule')
        }();

        if ((typeof _ret2 === 'undefined' ? 'undefined' : _typeof(_ret2)) === "object") return _ret2.v;
      } else {
        // simple rule 

        var _cache$_id3 = cache[_id];
        var type = _cache$_id3.type;
        var _style2 = _cache$_id3.style;

        styleBag[type] = _extends({}, styleBag[type] || {}, _style2);
        hasLabels && labels.push((_style2.label || '`' + _id) + ('' + (type !== '_' ? ':' + type : ''))); // todo - match 'add()'s original label
        return;
        // not too bad 
      }
    } else {
      // plain style 
      styleBag._ = _extends({}, styleBag._ || {}, rule);
      hasLabels && labels.push('{…}');
    }
  });

  // todo - remove label from merged styles? unclear. 

  var id = (0, _hash2.default)(mergeLabel + JSON.stringify(mediaBag) + JSON.stringify(styleBag)).toString(36); // todo - predictable order
  // make a merged label
  var label = hasLabels ? '' + (mergeLabel ? mergeLabel + '= ' : '') + (labels.length ? labels.join(' + ') : '') : ''; // yuck 

  if (!cache[id]) {
    cache[id] = _extends({ bag: styleBag, id: id, label: label }, Object.keys(mediaBag).length > 0 ? { media: mediaBag } : {});
    Object.keys(styleBag).forEach(function (type) {
      appendSheetRule(cssrule(id, type, styleBag[type]));
    });

    Object.keys(mediaBag).forEach(function (expr) {
      var css = Object.keys(mediaBag[expr]).map(function (type) {
        return cssrule(id, type, mediaBag[expr][type]);
      }).join('\n');

      appendSheetRule('@media ' + expr + ' { ' + css + ' }');
    });
  }
  return _defineProperty({}, 'data-css-' + id, label);
}

// this one's for media queries 
// they cannot be merged with other queries 
// todo - we should test whether the query is valid and give dev feedback 
function media(expr) {
  for (var _len3 = arguments.length, rules = Array(_len3 > 1 ? _len3 - 1 : 0), _key3 = 1; _key3 < _len3; _key3++) {
    rules[_key3 - 1] = arguments[_key3];
  }

  if (rules.length > 1) {
    return media(expr, merge.apply(undefined, rules));
  } // todo - iterate yourself instead 
  var rule = rules[0];
  // test if valid media query
  if (isRule(rule)) {
    var id = idFor(rule);

    if (cache[id].bag) {
      var _ret4 = function () {
        // merged rule       
        // todo - test if any media rules in this merged rule, throw if so 
        if (cache[id].media) {
          throw new Error('cannot apply a media rule onto another');
        }
        var bag = cache[id].bag;

        var newId = (0, _hash2.default)(expr + id).toString(36);
        var label = hasLabels ? '*mq [' + cache[id].label + ']' : '';

        if (!cache[newId]) {
          var cssRules = Object.keys(bag).map(function (type) {
            return cssrule(newId, type, bag[type]);
          });
          appendSheetRule('@media ' + expr + ' { ' + cssRules.join('\n') + ' }');
          cache[newId] = { expr: expr, rule: rule, id: newId, label: label };
        }

        return {
          v: _defineProperty({}, 'data-css-' + newId, label)
        };
        // easy 
      }();

      if ((typeof _ret4 === 'undefined' ? 'undefined' : _typeof(_ret4)) === "object") return _ret4.v;
    } else if (cache[id].expr) {
      // media rule
      throw new Error('cannot apply a media rule onto another');
    } else {
      // simple rule
      var _newId = (0, _hash2.default)(expr + id).toString(36);
      var _label = hasLabels ? '*mq ' + (cache[id].style.label || '`' + id) : '';

      if (!cache[_newId]) {
        appendSheetRule('@media ' + expr + ' { ' + cssrule(_newId, cache[id].type, cache[id].style) + ' }');
        cache[_newId] = { expr: expr, rule: rule, id: _newId, label: _label };
      }

      return _defineProperty({}, 'data-css-' + _newId, _label);
      // easier 
    }
  } else {
    // a plain style 
    var _style3 = rule;
    var _newId2 = styleHash(expr, _style3);
    var _label2 = hasLabels ? '*mq ' + (_style3.label || '') : '';
    if (!cache[_newId2]) {
      appendSheetRule('@media ' + expr + ' { ' + cssrule(_newId2, '_', _style3) + ' }');
      cache[_newId2] = { expr: expr, style: _style3, id: _newId2, label: _label2 };
    }
    return _defineProperty({}, 'data-css-' + _newId2, _label2);
  }
}

/**** live media query labels ****/

// simplest implementation -
// cycle through the cache, and for every media query
// find matching elements and update the label 
function updateMediaQueryLabels() {
  Object.keys(cache).forEach(function (id) {
    var expr = cache[id].expr;

    if (expr && hasLabels && window.matchMedia) {
      (function () {
        var els = document.querySelectorAll('[data-css-' + id + ']');
        var match = window.matchMedia(expr).matches ? '✓' : '✕';
        var regex = /^(✓|✕|\*)mq/;
        [].concat(_toConsumableArray(els)).forEach(function (el) {
          return el.setAttribute('data-css-' + id, el.getAttribute('data-css-' + id).replace(regex, match + 'mq'));
        });
      })();
    }
  });
}

// saves a reference to the loop we trigger 
var interval = void 0;

function trackMediaQueryLabels() {
  var bool = arguments.length <= 0 || arguments[0] === undefined ? true : arguments[0];
  var period = arguments.length <= 1 || arguments[1] === undefined ? 2000 : arguments[1];

  if (bool) {
    if (interval) {
      console.warn('already tracking labels, call trackMediaQueryLabels(false) to stop'); // eslint-disable-line no-console 
      return;
    }
    interval = setInterval(function () {
      return updateMediaQueryLabels();
    }, period);
  } else {
    clearInterval(interval);
    interval = null;
  }
}

// in dev mode, start this up immediately 
if (isDev && isBrowser) {
  trackMediaQueryLabels(true);
  // todo - make sure hot loading isn't broken
  // todo - clearInterval on browser close  
}

// we don't go all out for fonts as much, giving a simple font loading strategy 
// use a fancier lib if you need moar power
function fontFace(font) {
  var id = (0, _hash2.default)(JSON.stringify(font)).toString(36);
  if (!cache[id]) {
    cache[id] = { id: id, family: font.fontFamily, font: font };
    // todo - crossbrowser 
    appendSheetRule('@font-face { ' + (0, _CSSPropertyOperations.createMarkupForStyles)(font) + '}');
  }
  return font.fontFamily;
}

// we can add keyframes in a similar manner, but still generating a unique name 
// for including in styles. this gives us modularity, but still a natural api 
function keyframes(name, kfs) {
  if (typeof name !== 'string') {
    kfs = name;
    name = 'animate';
  }
  var id = (0, _hash2.default)(name + JSON.stringify(kfs)).toString(36);
  if (!cache[id]) {
    (function () {
      cache[id] = { id: id, name: name, kfs: kfs };
      var inner = Object.keys(kfs).map(function (kf) {
        return kf + ' { ' + (0, _CSSPropertyOperations.createMarkupForStyles)(prefixes(kfs[kf])) + '}';
      }).join('\n');

      ['-webkit-', '-moz-', '-o-', ''].forEach(function (prefix) {
        return appendSheetRule('@' + prefix + 'keyframes ' + (name + '_' + id) + ' { ' + inner + '}');
      });
    })();
  }
  return name + '_' + id;
}

/*** helpers for web components ***/
// https://github.com/threepointone/glamor/issues/16

function cssFor() {
  for (var _len4 = arguments.length, rules = Array(_len4), _key4 = 0; _key4 < _len4; _key4++) {
    rules[_key4] = arguments[_key4];
  }

  var ids = rules.reduce(function (o, r) {
    return o[idFor(r)] = true, o;
  }, {});
  var css = [].concat(_toConsumableArray(styleSheet.cssRules)).map(function (_ref5) {
    var cssText = _ref5.cssText;

    var regex = /\[data\-css\-([a-zA-Z0-9]+)\]/gm;
    var match = regex.exec(cssText);

    if (match && ids[match[1]]) {
      return cssText;
    }
  }).join('\n');
  return css;
}

function attribsFor() {
  for (var _len5 = arguments.length, rules = Array(_len5), _key5 = 0; _key5 < _len5; _key5++) {
    rules[_key5] = arguments[_key5];
  }

  var htmlAttributes = rules.map(function (rule) {
    idFor(rule); // throwaway check for rule 
    var key = Object.keys(rule)[0],
        value = rule[key];
    return key + '="' + (value || '') + '"';
  }).join(' ');

  return htmlAttributes;
}

/**** serverside stuff ****/

// the api's copied from aphrodite, with 1 key difference 
// we include *all* the css generated by the app 
// to optimize to only include generated styles on the pages 
// use renderStaticOptimized
function renderStatic(fn) {
  var optimized = arguments.length <= 1 || arguments[1] === undefined ? false : arguments[1];

  var html = fn();
  if (html === undefined) {
    throw new Error('did you forget to return from renderToString?');
  }
  var rules = [].concat(_toConsumableArray(styleSheet.cssRules)),
      css = rules.map(function (r) {
    return r.cssText;
  }).join('\n');
  if (optimized) {
    var _ret7 = function () {
      // parse out ids from html
      // reconstruct css/rules/cache to pass

      var o = { html: html, cache: {}, css: '' };
      var regex = /data\-css\-([a-zA-Z0-9]+)=/gm;
      var match = void 0,
          ids = [];
      while ((match = regex.exec(html)) !== null) {
        ids.push(match[1]);
      }
      ids.forEach(function (id) {
        o.cache[id] = cache[id];

        // todo - add fonts / animations
        o.css += rules.map(function (x) {
          return x.cssText;
        }).filter(function (r) {
          return new RegExp('\\[data-css-' + id + '\\]').test(r);
        }).join('\n') + '\n';
      });
      return {
        v: o
      };
    }();

    if ((typeof _ret7 === 'undefined' ? 'undefined' : _typeof(_ret7)) === "object") return _ret7.v;
  }
  return { html: html, cache: cache, css: css };
}

function renderStaticOptimized(fn) {
  return renderStatic(fn, true);
}

function rehydrate(c) {
  // load up cache
  cache = _extends({}, cache, c);
  // assume css loaded separately
}